#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2017-12-28
#
# Monitor status of Cisco VPN tunnel phase 1 and 2
#
# 2018-01-10: added handling for tunnel not found
# 2018-01-23: removed unnecessary counters
# 2018-02-15: removed ipsec tunnel status, changed ike ipv4 check
# 2018-02-16: readded tunnel alias
# 2018-07-11: added parameter for missing IPSec SA, changed 'parsed' to use peer ip as index
# 2021-08-03: rewritten for CMK 2.0
# 2021-11-03: fix negative vpn active tine
# 2022-01-19: added workaround for not matching IKE_OID_end and cipSecTunIkeTunnelIndexfor not matching
#             IKE_OID_end and cipSecTunIkeTunnelIndex, try to match IPSec nad IKE sa by remote address
# 2022-04-01: changed IPSec SA count output to check levels
# 2022-12-17: check input values for isdigit()
# 2023-06-07: fixed ValueError (not enough values to unpack (expected 10, got 4)) in parse function (ipsec_tunnel_entry)
# 2023-06-07: fixed ValueError (not enough values to unpack (expected 18, got 4)) in parse function (ike_tunnel_entry)
#

# snmpwalk sample
#
#
import time
from dataclasses import dataclass
from typing import List, Dict, Optional

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Service,
    Result,
    check_levels,
    State,
    SNMPTree,
    contains,
    OIDEnd,
    get_rate,
    GetRateError,
    get_value_store,
    IgnoreResultsError,
    Metric,
    render,
    all_of,
    exists,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    DiscoveryResult,
    CheckResult,
    StringTable,
)


@dataclass
class IpsecSa:
    sa_count: int
    ike_tunnel_alive: int
    active_time: int
    hc_in_octets: int
    in_pkts: int
    in_drop_pkts: int
    hc_out_octets: int
    out_pkts: int
    out_drop_pkts: int
    tun_remote_addr: str


@dataclass
class IkeSa:
    local_addr: str
    remote_addr: str
    active_time: int
    in_octets: int
    in_pkts: int
    in_drop_pkts: int
    out_octets: int
    out_pkts: int
    out_drop_pkts: int
    status: int
    nego_mode: int
    ipsec_summary: IpsecSa


###########################################################################
#
#  Helper functions
#
###########################################################################

def _ikepeertype(st: int) -> str:
    name = {
        1: 'ipAddrPeer',
        2: 'namePeer',
    }
    return name.get(st, f'unknown ({st})')


def _ikenegomode(st: int) -> str:
    name = {
        1: 'main',
        2: 'aggressive',
        3: 'IKEv2 main([3]?)'
    }
    return name.get(st, f'unknown ({st})')


def _tunnelstatus(st: int) -> str:
    name = {
        1: 'active',
        2: 'destroy',
    }
    return name.get(st, f'unknown ({st})')


def _cisco_vpn_tunnel_render_ipv4_address(bytestring):
    return '.'.join([f'{ord(m)}' for m in bytestring])


def _get_ipsec_sa_by_remote_address(ike_remote_address: str, ipsec_sa_summary: Dict[str, IpsecSa]) -> Optional[IpsecSa]:
    found_ipsec_sa = False
    ipsec_sa = IpsecSa(0, 0, 0, 0, 0, 0, 0, 0, 0, '')
    for ipsec_sa_to_check in ipsec_sa_summary.values():
        if ipsec_sa_to_check.tun_remote_addr == ike_remote_address:
            found_ipsec_sa = True
            ipsec_sa.sa_count += 1
            ipsec_sa.hc_in_octets += ipsec_sa_to_check.hc_in_octets
            ipsec_sa.in_pkts += ipsec_sa_to_check.in_pkts
            ipsec_sa.in_drop_pkts += ipsec_sa_to_check.in_drop_pkts
            ipsec_sa.hc_out_octets += ipsec_sa_to_check.hc_out_octets
            ipsec_sa.out_pkts += ipsec_sa_to_check.out_pkts
            ipsec_sa.out_drop_pkts += ipsec_sa_to_check.out_drop_pkts
            if ipsec_sa_to_check.active_time > ipsec_sa.active_time:
                ipsec_sa.active_time = ipsec_sa_to_check.active_time
    if found_ipsec_sa:
        return ipsec_sa


###########################################################################
#
#  DATA Parser function
#
###########################################################################


def parse_cisco_vpn_tunnel(string_table: List[StringTable]) -> Dict[str, IkeSa]:
    ipsec_sa_summary: Dict[str, IpsecSa] = {}
    vpntunnel = {}
    ike_tunnel_entry, ipsec_tunnel_entry = string_table

    # summarize IPSec SAs, ASSUMPTION: except for counters all SA attributes are identical per IKE index
    for entry in ipsec_tunnel_entry:
        try:
            ike_tunnel_index, ike_tunnel_alive, tun_remote_addr, active_time, hc_in_octets, in_pkts, in_drop_pkts, \
            hc_out_octets, out_pkts, out_drop_pkts = entry
        except ValueError:
            continue

        if ike_tunnel_index.isdigit():
            ipsec_sa = ipsec_sa_summary.setdefault(
                ike_tunnel_index,
                IpsecSa(0, 0, 0, 0, 0, 0, 0, 0, 0, tun_remote_addr)
            )
            ipsec_sa.sa_count += 1
            ipsec_sa.hc_in_octets += int(hc_in_octets) if hc_in_octets.isdigit() else 0
            ipsec_sa.in_pkts += int(in_pkts) if in_pkts.isdigit() else 0
            ipsec_sa.in_drop_pkts += int(in_drop_pkts) if in_drop_pkts.isdigit() else 0
            ipsec_sa.hc_out_octets += int(hc_out_octets) if hc_out_octets.isdigit() else 0
            ipsec_sa.out_pkts += int(out_pkts) if out_pkts.isdigit() else 0
            ipsec_sa.out_drop_pkts += int(out_drop_pkts) if out_drop_pkts.isdigit() else 0
            if active_time.isdigit and (int(active_time) // 100 > ipsec_sa.active_time):
                ipsec_sa.active_time = int(active_time) // 100

    # IKE SA
    for entry in ike_tunnel_entry:
        try:
            index, local_type, local_value, local_addr, local_name, remote_type, remote_value, remote_addr, remote_name, \
            active_time, in_octets, in_pkts, in_droppkts, out_octets, out_pkts, out_droppkts, status, \
            nego_mode = entry
        except ValueError:
            continue

        if index.isdigit():
            save_remote_addr = remote_addr
            remote_addr = _cisco_vpn_tunnel_render_ipv4_address(remote_addr)
            if remote_addr.split('.') != 4:
                remote_addr = remote_value
            if len(remote_addr.split('.')) == 4:
                ike_sa = IkeSa(
                    local_addr=_cisco_vpn_tunnel_render_ipv4_address(local_addr),
                    remote_addr=remote_addr,
                    active_time=int(active_time) // 100 if active_time.isdigit() else 0,
                    in_octets=int(in_octets) if in_octets.isdigit() else 0,
                    in_pkts=int(in_pkts) if in_pkts.isdigit() else 0,
                    in_drop_pkts=int(in_droppkts) if in_droppkts.isdigit() else 0,
                    out_octets=int(out_octets) if out_octets.isdigit() else 0,
                    out_pkts=int(out_pkts) if out_pkts.isdigit() else 0,
                    out_drop_pkts=int(out_droppkts) if out_droppkts.isdigit() else 0,
                    status=int(status) if status.isdigit() else 0,
                    nego_mode=int(nego_mode) if nego_mode.isdigit() else 0,
                    # if/else is workaround for not matching IKE_OID_end and cipSecTunIkeTunnelIndex
                    # try to match IPSec sa by remote address
                    ipsec_summary=ipsec_sa_summary.get(index) if ipsec_sa_summary.get(
                        index) is not None else _get_ipsec_sa_by_remote_address(save_remote_addr, ipsec_sa_summary)
                )
                vpntunnel.update({remote_addr: ike_sa})

    return vpntunnel


###########################################################################
#
#  Inventory function
#
###########################################################################


def discovery_cisco_vpn_tunnel(params, section: Dict[str, IkeSa]) -> DiscoveryResult:
    discover_aggressive_mode = params['discover_aggressive_mode']
    for cikeTunRemoteAddr in section.keys():
        if section[cikeTunRemoteAddr].nego_mode != 2:
            yield Service(item=cikeTunRemoteAddr)
        elif discover_aggressive_mode:
            yield Service(item=cikeTunRemoteAddr)


###########################################################################
#
#  Check function
#
###########################################################################


def check_cisco_vpn_tunnel(item, params, section: Dict[str, IkeSa]) -> CheckResult:
    tunnel_not_found_state = params['state']
    missing_ipsec_sa_state = params['missing_ipsec_sa_state']

    for tunnel_ip, tunnel_alias, not_found_state, ipsec_sa_state in params['tunnels']:
        if item == tunnel_ip:
            yield Result(state=State.OK, summary=f'[{tunnel_alias}]')
            tunnel_not_found_state = not_found_state
            missing_ipsec_sa_state = ipsec_sa_state

    try:
        tunnel = section[item]
    except KeyError:
        yield Result(state=State(tunnel_not_found_state), summary='VPN Tunnel not found in SNMP data')
        return

    # check if tunnel lifetime < 0, happens on asa failover
    if tunnel.active_time < 0:
        tunnel.active_time = tunnel.active_time * -1

    yield from check_levels(
        value=tunnel.active_time,
        label='IKE uptime',
        render_func=render.timespan,
        metric_name='cisco_vpn_tunnel_cikeTunActiveTime'
    )

    yield Result(state=State.OK, notice=f'IKE Status: {_tunnelstatus(tunnel.status)}')
    yield Result(state=State.OK, notice=f'Tunnel address local: {tunnel.local_addr}')
    yield Result(state=State.OK, notice=f'Tunnel address remote : {tunnel.remote_addr}')
    yield Result(state=State.OK, notice=f'Negotiation mode : {_ikenegomode(tunnel.nego_mode)}')

    now_time = time.time()
    value_store = get_value_store()
    rate_item = item.replace(' ', '_').replace(':', '_')
    raise_ingore_res = False

    # convert to octets/packets per second
    for key, value in [
        ('cikeTunInOctets', tunnel.in_octets),
        ('cikeTunOutOctets', tunnel.out_octets),
        ('cikeTunInPkts', tunnel.in_pkts),
        ('cikeTunOutPkts', tunnel.out_pkts),
        ('cikeTunInDropPkts', tunnel.in_drop_pkts),
        ('cikeTunOutDropPkts', tunnel.out_drop_pkts),

    ]:
        try:
            value = get_rate(value_store, f'cisco_vpn_tunnel.{key}.{rate_item}', now_time, value, raise_overflow=False)
        except GetRateError:
            raise_ingore_res = True
            value = 0
        yield Metric(name=f'cisco_vpn_tunnel_{key}', value=value, boundaries=(0, None))

    if raise_ingore_res:
        raise IgnoreResultsError('Initializing counters')

    ipsecsummary: IpsecSa = tunnel.ipsec_summary
    if ipsecsummary is not None:

        yield from check_levels(
            label='IPSec uptime',
            value=ipsecsummary.active_time,
            render_func=render.timespan,
            metric_name='cisco_vpn_tunnel_cipSecTunActiveTime'
        )

        yield from check_levels(
            label='IPSec SAs',
            value=ipsecsummary.sa_count,
            metric_name='cisco_vpn_tunnel_ipsec_sa_count',
            render_func=lambda v: f'{int(v)}',
        )
        # yield Result(state=State.OK, summary=f'SAs: {ipsecsummary.sa_count}')
        ipsec_in_octets = 0
        ipsec_out_octets = 0
        # convert to octets/packets per second
        for key, value in [
            ('cipSecTunHcInOctets', ipsecsummary.hc_in_octets),
            ('cipSecTunHcOutOctets', ipsecsummary.hc_out_octets),
            ('cipSecTunInPkts', ipsecsummary.in_pkts),
            ('cipSecTunOutPkts', ipsecsummary.out_pkts),
            ('cipSecTunInDropPkts', ipsecsummary.in_drop_pkts),
            ('cipSecTunOutDropPkts', ipsecsummary.out_drop_pkts),
        ]:
            try:
                value = get_rate(value_store, f'cisco_vpn_tunnel.{key}.{rate_item}',
                                 now_time, value, raise_overflow=False)
            except GetRateError:
                raise_ingore_res = True
                value = 0
            yield Metric(name=f'cisco_vpn_tunnel_{key}', value=value, boundaries=(0, None))
            if key == 'cipSecTunHcInOctets':
                ipsec_in_octets = value
            elif key == 'cipSecTunHcOutOctets':
                ipsec_out_octets = value

        yield from check_levels(
            label='In',
            value=ipsec_in_octets,
            render_func=render.networkbandwidth,
        )
        yield from check_levels(
            label='Out',
            value=ipsec_out_octets,
            render_func=render.networkbandwidth,
        )

        if raise_ingore_res:
            raise IgnoreResultsError('Initializing counters')
    else:
        yield Result(state=State(missing_ipsec_sa_state), notice='No IPSec sa found')


###########################################################################
#
#  Check info
#
###########################################################################


register.snmp_section(
    name='cisco_vpn_tunnel',
    parse_function=parse_cisco_vpn_tunnel,
    fetch=[
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.171.1.2.3.1',  #
            oids=[
                OIDEnd(),  # TunnelIndex
                '2',  # cikeTunLocalType
                '3',  # cikeTunLocalValue
                '4',  # cikeTunLocalAddr
                '5',  # cikeTunLocalName
                '6',  # cikeTunRemoteType
                '7',  # cikeTunRemoteValue
                '8',  # cikeTunRemoteAddr
                '9',  # cikeTunRemoteName
                '16',  # cikeTunActiveTime
                '19',  # cikeTunInOctets
                '20',  # cikeTunInPkts
                '21',  # cikeTunInDropPkts
                '27',  # cikeTunOutOctets
                '28',  # cikeTunOutPkts
                '29',  # cikeTunOutDropPkts
                '35',  # cikeTunStatus
                '10',  # cikeTunNegoMode
            ]
        ),
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.171.1.3.2.1',  # CISCO-IPSEC-FLOW-MONITOR-MIB::cipSecTunnelEntry
            oids=[
                '2',  # cipSecTunIkeTunnelIndex/ ike tunnel index
                '3',  # cipSecTunIkeTunnelAlive
                '5',  # cipSecTunRemoteAddr (needed as workaround if if cipSecTunIkeTunnelIndex and IKE OID don't match)
                '10',  # cipSecTunActiveTime
                '27',  # cipSecTunHcInOctets
                '32',  # cipSecTunInPkts
                '33',  # cipSecTunInDropPkts
                '40',  # cipSecTunHcOutOctets
                '45',  # cipSecTunOutPkts
                '46',  # cipSecTunOutDropPkts
            ]
        ),
    ],
    detect=all_of(
        contains('.1.3.6.1.2.1.1.1.0', 'Cisco'),
        exists('.1.3.6.1.4.1.9.9.171.1.2.3.1.2.*')  # CISCO-IPSEC-FLOW-MONITOR-MIB::cikeTunnelEntry
    ),
)

register.check_plugin(
    name='cisco_vpn_tunnel',
    service_name='VPN Tunnel %s',
    discovery_function=discovery_cisco_vpn_tunnel,
    discovery_ruleset_name='discovery_cisco_vpn_tunnel',
    discovery_default_parameters={
        'discover_aggressive_mode': False
    },
    check_function=check_cisco_vpn_tunnel,
    check_default_parameters={
        'state': 3,  # default state for tunnel not found
        'missing_ipsec_sa_state': 1,
        'tunnels': [],  # list of tunnel specific not found states
                        # ('<ip-address>', '<alias>', <not_found_state>, <no_ipsec_sa>)
    },
    check_ruleset_name='cisco_vpn_tunnel',
)
